#include "Query.h"
#include "MismatchQuery.h"
#include "IndelQuery.h"
#include "RotationQuery.h"

#include <sstream>
#include <iostream>
using std::cout;
using std::endl;
using std::cerr;
#include <string>
using std::string;
#include <ctime>
#include <cstring>
#include <getopt.h>

#ifdef PARALLEL_SUPPORT
#include <omp.h>
#endif

/**
 * Definitions for parsing command line options
 */
enum align_mode_t { mode_undef, mode_mismatch, mode_rotation,
                    mode_indel, mode_gap, mode_matepair };

enum parameter_t { long_opt_all = 256, long_opt_maxgap,
                   long_opt_minprefix, long_opt_skip, long_opt_nreads,
                   long_opt_debug, long_opt_recursion };

void print_usage(char const *name)
{
    cerr << "usage: " << name << " [options] <index> <input>" << endl
         << "Check README or `" << name << " --help' for more information." << endl;
}

void print_help(char const *name)
{
    cerr << "usage: " << name << " [options] <index> <input>" << endl
         << endl
         << "Input files:" << endl
         << "  <index>   Index filename, see `bwt_make --help' for more information" << endl
         << "            about constructing indexes." << endl
         << endl
         << "  <input>   Filename for input reads, give \"-\" as a filename to read from " << endl
         << "            the standard input. Default file format is one-read-per-line." << endl
         << "            For FASTQ or FASTA inputs, use options -q and -f, respectively." << endl
         << endl
         << "Input options (default input format is one-read-per-line):" << endl
         << " -q, --fastq                 Input reads in FASTQ format." << endl
         << " -f, --fasta                 Input reads in FASTA format." << endl
         << " --skip <int>                Skip first <int> reads from input (default is 0)." << endl
         << " --nreads <int>              Align <int> reads from input (after skipping)" << endl
         << "                             (default is all reads)." << endl
         << " -c, --color                 Reads are in SOLiD color codes." << endl
         << " -t <int>, --truncate <int>  Truncate <int> characters from beginning and end" << endl
         << "                             of all patterns." << endl
         << endl
         << "Output options (default output format is tab-delimited, see README):" << endl
         << " -o <name>, --output <name>  Output filename, default is the standard output." << endl
         << " -s, --sam                   Output in SAMTools [2] format." << endl
         << " -r <int>, --report <int>    Report at most <int> alignments per read " << endl
         << "                             (default is one)." << endl
         << " --all                       Report all occurrences (can be extremely slow)." << endl
         << " -v, --verbose               Verbose mode, uses the standard error output." << endl
         << endl
         << "Basic alignment modes (ChIP-seq, resequencing):" << endl
         << " -k <int>, --mismatch <int>  Allow <int> mismatches in the aligment." << endl
         << " -R <int>, --rotation <int>  Allow <int> mismatches (as in -k) but use a" << endl
         << "                             rotation index; see README how to build it." << endl
         << " -i <int>, --indel <int>     Allow <int> mismatches, insertions or deletions" << endl
         << "                             in the aligment." << endl
         << endl
         << "Alignment modes including a gap (RNA-seq):" << endl
         << " -g <int>, --gap <int>       Single-read search with <int> mismatches and " << endl
         << "                             one gap in unknown position." << endl
         << " -p <int>, --mate-pair <int> Search mate-pairs with <int> mismatches and " << endl
         << "                             one gap between reads. Requires two input files " << endl
         << "                             that contain mate-pairs in the same order." << endl
         << endl
         << "Gap alignment options (for modes -g and -p):" << endl
         << " --max-gap <int>             Maximum gap length (default: 500)" << endl
         << " --min-prefix <int>          Minimum number of characters before the gap" << endl
         << "                             (default is half of the pattern length)" << endl
         << endl
         << "General options:" << endl
         << " -h, --help                  Display command line options." << endl
         << " --recursion <int>           Maximum number of recursive calls per search " << endl
         << "                             (default: 999999999)." << endl;
#ifdef PARALLEL_SUPPORT
    cerr << " -P <int>, --parallel <int>  Number of parallel threads to use (default is one," << endl
         << "                             give argument 0 to use all available cores)." << endl;
#else
    cerr << " -P <int>, --parallel <int>  Parallel computation requires OpenMP, see README." << endl;
#endif
}

int atoi_min(char const *value, int min, char const *parameter, char const *name)
{
    std::istringstream iss(value);
    int i;
    char c;
    if (!(iss >> i) || iss.get(c))
    {
        cerr << "readaligner: argument of " << parameter << " must be of type <int>, and greater than or equal to " << min << endl
             << "Check README or `" << name << " --help' for more information." << endl;
        std::exit(1);
    }

    if (i < min)
    {
        cerr << "readaligner: argument of " << parameter << " must be greater than or equal to " << min << endl
             << "Check README or `" << name << " --help' for more information." << endl;
        std::exit(1);
    }
    return i;
}

int main(int argc, char **argv) 
{
    MyersEditDistanceIncremental::initMyersFourRussians();

    /**
     * Parse command line parameters
     */
    if (argc <= 1)
    {
        print_usage(argv[0]);
        return 0;
    }

    bool incremental = true;
    bool reverse = true;
    bool verbose = false;
    bool color = false;
    bool debug = false;
    unsigned maxk = ~0u;
    int maxgap = 500;
    int gapstart = -1;
    unsigned rlimit = 999999999u;
    unsigned truncate = 0;
    unsigned report = 1;
    unsigned skippatterns = 0;
    unsigned numberofpatterns = 0;
#ifdef PARALLEL_SUPPORT
    unsigned parallel = 1;
#endif
    string outputfile = "";

    align_mode_t mode = mode_undef;
    OutputWriter::output_format_t output = OutputWriter::output_tabs;
    InputReader::input_format_t input = InputReader::input_lines;
    
    static struct option long_options[] =
        {
            {"fastq",     no_argument,       0, 'q'},
            {"fasta",     no_argument,       0, 'f'},
            {"skip",      required_argument, 0, long_opt_skip},
            {"nreads",    required_argument, 0, long_opt_nreads},
            {"color",     no_argument,       0, 'c'},
            {"truncate",  required_argument, 0, 't'},
            {"output",    required_argument, 0, 'o'},
            {"sam",       no_argument,       0, 's'},
            {"report",    required_argument, 0, 'r'},
            {"all",       no_argument,       0, long_opt_all},
            {"verbose",   no_argument,       0, 'v'},
            {"mismatch",  required_argument, 0, 'k'},
            {"rotation",  required_argument, 0, 'R'},
            {"indel",     required_argument, 0, 'i'},
            {"gap",       required_argument, 0, 'g'},
            {"mate-pair", required_argument, 0, 'p'},
            {"max-gap",   required_argument, 0, long_opt_maxgap},
            {"min-prefix",required_argument, 0, long_opt_minprefix},
            {"help",      no_argument,       0, 'h'},
            {"recursion", required_argument, 0, long_opt_recursion},
            {"parallel",  required_argument, 0, 'P'},
            {"debug",     no_argument,       0, long_opt_debug},
            {0, 0, 0, 0}
        };
    int option_index = 0;
    int c;
    while ((c = getopt_long(argc, argv, "qfct:o:sr:vk:R:i:g:p:hP:",
                                 long_options, &option_index)) != -1) 
    {
        switch(c) 
        {
        case 'q': 
            input = InputReader::input_fastq; break;
        case 'f':
            input = InputReader::input_fasta; break;
        case long_opt_skip:
            skippatterns = atoi_min(optarg, 0, "--skip", argv[0]); break;
        case long_opt_nreads:
            numberofpatterns = atoi_min(optarg, 1, "--nreads", argv[0]); break;
        case 'c':
            color = true; break;
        case 't':
            truncate = atoi_min(optarg, 0, "-t, --truncate", argv[0]); break;
        case 'o':
            outputfile = string(optarg); break;
        case 's':
            output = OutputWriter::output_sam; break;
        case 'r':
            if (report == ~0u)
            {
                cerr << "readaligner: parameters --all and -r are mutually exclusive." << endl;
                return 1;
            }
            report = atoi_min(optarg, 1, "-r, --report", argv[0]); break;
        case long_opt_all:
            if (report != 1)
            {
                cerr << "readaligner: parameters --all and -r are mutually exclusive." << endl;
                return 1;
            }
            report = ~0u; break;
        case 'v':
            verbose = true; break;
        case 'k':
        case 'R':
        case 'i':            
        case 'g':
        case 'p':
            if (mode != mode_undef)
            {
                cerr << "readaligner: parameters -k, -R, -i, -g and -p are mutually exclusive." << endl;
                return 1;
            }
            switch (c)
            {
            case 'k':
                mode = mode_mismatch;
                maxk = atoi_min(optarg, 0, "-k, --mismatch", argv[0]); break;
            case 'R':
                mode = mode_rotation;
                maxk = atoi_min(optarg, 0, "-R, --rotation", argv[0]); break;
            case 'i':
                mode = mode_indel;
                maxk = atoi_min(optarg, 0, "-i, --indel", argv[0]); break;
            case 'g':
                mode = mode_gap;
                maxk = atoi_min(optarg, 0, "-g, --gap", argv[0]); break;
            case 'p':
                mode = mode_matepair;
                maxk = atoi_min(optarg, 0, "-p, --mate-pair", argv[0]); break;
            default:
                abort();
            }
            break;
        case long_opt_maxgap:
            maxgap = atoi_min(optarg, 1, "--max-gap", argv[0]); break;
        case long_opt_minprefix:
            gapstart = atoi_min(optarg, 0, "--min-prefix", argv[0]); break;
        case 'h':
            print_help(argv[0]);
            return 0;
        case long_opt_recursion:
            rlimit = atoi_min(optarg, 1, "--recursion", argv[0]); break;
        case 'P':
#ifdef PARALLEL_SUPPORT
            parallel = atoi_min(optarg, 0, "-P, --parallel", argv[0]); break;
#else
            cerr << "readaligner: Parallel processing not currently available!" << endl 
                 << "Please recompile with parallel support; see README for more information." << endl;
            return 1;
#endif
        case long_opt_debug:
            debug = true; break;
        case '?':
            print_usage(argv[0]);
            return 1;
        default:
            print_usage(argv[0]);
            std::abort ();
        }
    }

    if (mode == mode_undef)
    {
        cerr << "readaligner: one of the parameters -k, -R, -i, -g or -p must be given" << endl;
        print_usage(argv[0]);
        return 1;
    }

    // Sanity checking
    if (maxk == ~0u)
    {
        cerr << "readaligner: maxk was not set, internal error." << endl;
        std::abort();
    }

    // Parse filenames
    if (argc - optind != 2)
    {
        cerr << "readaligner: missing input files!" << endl;
        print_usage(argv[0]);
        return 1;
    }

    string indexfile = string(argv[optind++]);
    string patternfile = string(argv[optind++]);
    
    /**
     * Initialize shared data structures
     */
    if (verbose) cerr << "Loading index " << indexfile << endl;
    TextCollection *tc = 0; 
    TextCollection *tcr = 0;
    switch (mode)
    {
    case mode_indel:
    case mode_gap:
    case mode_matepair:
        tc = TextCollection::load(indexfile);
        break;
    case mode_mismatch:
        tc = TextCollection::load(indexfile);
        {
            string indexreversefile = indexfile + TextCollection::REVERSE_EXTENSION;
            if (verbose) cerr << "Loading reverse index " << indexreversefile << endl;
            tcr = TextCollection::load(indexreversefile);
            if (!tcr)
            {
                // Check without suffix
                std::size_t found = indexfile.rfind(".");
                if (found != string::npos)
                    tcr = TextCollection::load(indexfile.substr(0,found) + TextCollection::REVERSE_EXTENSION);
            }
            if (!tcr) 
            {
                cerr << "unable to read reverse index file " << indexreversefile << endl;
                return 1;
            }
        }
        break;
    case mode_rotation:
        {
            string rotationfile = indexfile + TextCollection::ROTATION_EXTENSION;
            if (verbose) cerr << "Loading rotation index " << rotationfile << endl;
            tc = TextCollection::load(rotationfile);
            if (!tc)
            {
                // Check without suffix
                std::size_t found = indexfile.rfind(".");
                if (found != string::npos)
                    tc = TextCollection::load(indexfile.substr(0,found) + TextCollection::ROTATION_EXTENSION);
            }
            if (!tc) 
            {
                cerr << "unable to read rotation index file " << rotationfile << endl;
                return 1;
            }
        }
        break;
    case mode_undef:
    default:
        abort();
        break;
    }

    // Sanity checks
    if (!tc) {
        cerr << "readaligner: could not read index file " << indexfile << endl;
        return 1;
    }
    if (tc->isColorCoded() && !color)
    {
        cerr << "readaligner: index " << indexfile << " is color coded, please use option -c, --color." << endl;
        return 1;
    }
    if (!tc->isColorCoded() && color)
    {
        cerr << "readaligner: option -c, --color was given, but the index " << indexfile 
             << " is not color coded!" << endl 
             << "Please rebuild the index using -c, --color option, see README for more information." << endl;
        return 1;
    }

    
    /**
     * Initialize shared input reader(s)
     */
    InputReader *inputr = InputReader::build(input, numberofpatterns, patternfile);
    assert(inputr != 0);
    if (!inputr->skip(skippatterns))
    {
        cerr << "unexpected end of file: unable to skip " << skippatterns << " reads from input!" << endl
             << "Please check README and make sure the input file format is valid." << std::endl;
        std::exit(1);
    }


    /**
     * Initialize shared output writer
     */
    OutputWriter *outputw = OutputWriter::build(output, outputfile);
    assert(outputw != 0);
    cerr << std::fixed;
    cerr.precision(2);

    // Shared counters
    unsigned total_found = 0;
    ulong total_occs = 0;
    time_t wctime = time(NULL);
    time_t wcrate = time(NULL);

#ifdef PARALLEL_SUPPORT
    if (parallel != 0)
    {
        if (verbose && parallel == 1) 
            cerr << "Using only one core (default setting, see option -P, --parallel)" << endl;
        if (verbose && parallel != 1)
            cerr << "Using " << parallel << " cores." << endl;
        omp_set_num_threads(parallel);
    }
    else
        if (verbose) cerr << "Using all " << omp_get_max_threads() << " cores available." << endl;
#pragma omp parallel 
#endif    
{
    Query *query = 0; // Private query instances

#ifdef PARALLEL_SUPPORT
#pragma omp critical (CONSTRUCT_QUERY)
#endif    
{
    /**
     * Initialize query
     */
    switch (mode)
    {
    case mode_undef:
        abort();
        break;
    case mode_mismatch:
        query = new MismatchQuery(tc, tcr, *outputw, verbose, incremental, reverse, report);
        if (verbose) cerr << "Align mode: k-mismatch (-k, --mismatch) with k = " << maxk << endl;
        break;
    case mode_rotation:
        query = new RotationQuery(tc, *outputw, verbose, incremental, reverse, report);
        if (verbose) cerr << "Align mode: k-mismatch using rotation index (-R, --rotation) with k = " << maxk << endl;
        break;
    case mode_indel:
        query = new IndelQuery(tc, *outputw, verbose, incremental, reverse, report);
        if (verbose) cerr << "Align mode: k-errors (-i, --indel) with k = " << maxk << endl;
        break;
/*    case mode_gap:
        query = new GapQuery(tc, *outputw, verbose, incremental, reverse, report, maxgap, gapstart);
        if (verbose) cerr << "Align mode: k-mismatch with a gap (-g, --gap), k = " << maxk << " and maxgap = " << maxgap << endl;
        break;
    case mode_matepair:
        query = new MatepairQuery(...);
        break;*/
    default:
        cerr << "error: Alignment mode not implemented yet?" << endl;
        abort();
        break;
    }
    assert(query != 0);
    // Other query settings
    query->setDebug(debug);
    query->setRecursionLimit(rlimit);

} // End of #pragma omp critical (CONSTRUCT_QUERY)

    /**
     * Main loop: iterate through input
     */
    Pattern::setColor(color);
    Pattern pattern;
    while (inputr->next(pattern)) 
    {
    	query->reinit();
        pattern.truncate(truncate);
        if (debug)
            cerr << "****************************************************************" << endl
                 << "Next pattern: " << pattern.getOrigPattern() << " (" << pattern.getName() << ")" << endl;

        unsigned k = maxk;
        if (incremental)
            k = 0;
        
        unsigned reported = 0;
        while (k <= maxk && reported < report)
        {
            query->align(&pattern, k, reported);
            ++k;
        }

        if (verbose && (pattern.getId()+1) % 10000 == 0)
        {
#ifdef PARALLEL_SUPPORT
#pragma omp critical (STDERR)
#endif
            {
                cerr << "after " << inputr->getCount() << " reads: total time " 
                     << std::difftime(time(NULL), wctime) << " s (" 
                     << std::difftime(time(NULL), wctime) / 3600 << " hours), approx " 
                     << std::difftime(time(NULL), wcrate) / 10 << " ms per read, " 
                     << "found " << total_found << " reads (" << (double)total_found*100/inputr->getCount()
                     << " %)" << endl;
                wcrate = time(NULL);
            }
        }

#ifdef PARALLEL_SUPPORT
#pragma omp atomic
        total_found += reported ? 1 : 0;
#pragma omp atomic
        total_occs += reported;
#else
        total_found += reported ? 1 : 0;
        total_occs += reported;
#endif
    }       

    delete query;
} // end of #pragma omp parallel

    if (verbose && inputr->getCount())
    {	
        cerr << "Number of reads: " << inputr->getCount() << " (skipped first " << skippatterns << ")" << endl
             << "Successfully aligned: " << total_found << " reads (" << (double)total_found*100/inputr->getCount()
             << " %)" << endl
             << "Number of reported alignments: " << total_occs << endl;
        cerr << "Wall-clock time: " << std::difftime(time(NULL), wctime) << " seconds (" 
             << std::difftime(time(NULL), wctime) / 3600 << " hours)" << endl;
        cerr << "Average wall-clock time per read: " << std::difftime(time(NULL), wctime) * 1000 / inputr->getCount() 
             << " milliseconds" << endl;
    }
    if (!inputr->getCount())
        cerr << "readaligner: unable to read input! Please check the input file format and README." << endl;

    delete inputr; 
    delete outputw;
    delete tc;
    delete tcr;
}
